@@ -114,10 +114,6 @@
     
     (sx-match tree
 
-      ((program . ,parts)
-       (sf "%% program\n\n")
-       (for-each ppxin parts))
-
       ((script-file . ,items)
        (for-each ppxin items))
 
@@ -192,8 +188,6 @@
 
       ((ident ,ident)
        (sf "~A" ident))
-      ((name ,name)
-       (sf "~A" name))
 
       ((qident . ,names)
        (sf "~A" (string-join names ".")))
